import numpy as np
import statsmodels.api as sm

from GHOST_standards import standard_parameters

class Data_Products(object):

    """Provides functions for calculation of data products"""

    @staticmethod
    def calculate_aod(new_wavelength, calculation_params, calculation_arr):
        """ Calculate aerosol optical depth (AOD) at a given wavelength using following the Ångström power law

            :param new_wavelength: numerical value of the wavelength at which the new component is computed for
            :type new_wavelength: int
            :param calculation_params: GHOST names of the input components 
            :type calculation_params: list
            :param calculation_arr: input component values of shape (N components, time)
            :type calculation_arr: numpy.ndarray
            :return: new component array
            :rtype: numpy.ndarray
        """

        #filter out input data outside valid bounds
        for calculation_param_ii, calculation_param in enumerate(calculation_params):
            out_of_bounds = np.where(np.logical_or(calculation_arr[calculation_param_ii, :] < standard_parameters[calculation_param]['extreme_lower_limit'], calculation_arr[calculation_param_ii, :] > standard_parameters[calculation_param]['extreme_upper_limit']))[0]
            calculation_arr[calculation_param_ii, out_of_bounds] = np.NaN

        #get indices for aod and angstrom exponent parameters 
        aod_param_indices = np.array([calculation_param_ii for calculation_param_ii, calculation_param in enumerate(calculation_params) if 'AOD_' in calculation_param])
        #ae_param_index = [calculation_param_ii for calculation_param_ii, calculation_param in enumerate(calculation_params) if 'AE_' in calculation_param][0]
        
        #get specific parameters / angstrom exponent for aod and angstrom exponent 
        aod_params = [calculation_params[aod_param_index] for aod_param_index in aod_param_indices]
        aod_arr = calculation_arr[aod_param_indices, :]
        #ae_param = calculation_params[ae_param_index]
        #ae_arr = calculation_arr[ae_param_index, :]

        #get numerical values from aod params
        aod_wavelengths = [int(aod_param.split('_')[-1][:-2]) for aod_param in aod_params]

        #calculate new AOD component (Sara's method)
        #factors = np.full(aod_arr.shape, np.NaN, dtype=np.float32)
        #for aod_wavelength_ii, aod_wavelength in enumerate(aod_wavelengths):
        #    factors[aod_wavelength_ii, :] = (aod_arr[aod_wavelength_ii, :]/(aod_wavelength**(-ae_arr[:])))*(new_wavelength**(-ae_arr[:]))            
        #new_aod = np.mean(factors, axis=0)

        #calculate new AOD component (logfit method)
        new_aod = np.full((aod_arr.shape[1]), np.NaN, dtype=np.float32)
        log_aod_wavelengths = np.log(aod_wavelengths) 
        log_aod_wavelengths = np.column_stack((log_aod_wavelengths, log_aod_wavelengths**2))   
        log_aod_wavelengths = sm.add_constant(log_aod_wavelengths)
        log_aod_arr = np.log(aod_arr)  
        new_wavelength_log = np.log(new_wavelength)
        #iterate through time points
        for time_ii in range(len(new_aod)):
            log_aod_vals = log_aod_arr[:,time_ii]
            #only proceed with calculation if have all values for calculation
            if np.isfinite(log_aod_vals).aall():
                model = sm.OLS(log_aod_vals, log_aod_wavelengths).fit()
                A0 = model.params[2]
                B0 = model.params[1]
                C0 = model.params[0]
                #model = LinearRegression().fit(log_aod_wavelengths, log_aod_vals)
                #A0 = model.coef_[1]
                #B0 = model.coef_[0] 
                #C0 = model.intercept_
                new_aod[time_ii] = np.exp(A0*new_wavelength_log**2 + B0*new_wavelength_log + C0)

        return new_aod

    @staticmethod
    def calculate_ae(calculation_params, calculation_arr):
        """ Calculate angstrom exponent (AE) using aerosol coefficients at different wavelengths

            :param calculation_params: GHOST names of the input components 
            :type calculation_params: list
            :param calculation_arr: input component values of shape (N components, time)
            :type calculation_arr: numpy.ndarray
            :return: new component array
            :rtype: numpy.ndarray
        """

        #filter out input data outside valid bounds
        for calculation_param_ii, calculation_param in enumerate(calculation_params):
            out_of_bounds = np.where(np.logical_or(calculation_arr[calculation_param_ii, :] < standard_parameters[calculation_param]['extreme_lower_limit'], calculation_arr[calculation_param_ii, :] > standard_parameters[calculation_param]['extreme_upper_limit']))[0]
            calculation_arr[calculation_param_ii, out_of_bounds] = np.NaN

        #get wavelengths from calculation params 
        wavelengths = [int(calculation_param.split('_')[-1][:-2]) for calculation_param in calculation_params]
        log_wavelengths = np.log(wavelengths)

        #calculate new AE component
        new_ae = np.full((calculation_arr.shape[1]), np.NaN, dtype=np.float32)
        #iterate through time points
        for time_ii in range(len(new_ae)):
            calculation_vals_log = np.log(calculation_arr[:,time_ii])
            if np.isfinite(calculation_vals_log).all():
                new_ae[time_ii] = -np.polyfit(log_wavelengths,calculation_vals_log,1)[0]

        return new_ae

    @staticmethod
    def calculate_bf(calculation_params, calculation_arr):
        """ Calculate backscatter fraction at a given wavelength

            :param calculation_params: GHOST names of the input components 
            :type calculation_params: list
            :param calculation_arr: input component values of shape (N components, time)
            :type calculation_arr: numpy.ndarray
            :return: new component array
            :rtype: numpy.ndarray
        """

        #filter out input data outside valid bounds
        for calculation_param_ii, calculation_param in enumerate(calculation_params):
            out_of_bounds = np.where(np.logical_or(calculation_arr[calculation_param_ii, :] < standard_parameters[calculation_param]['extreme_lower_limit'], calculation_arr[calculation_param_ii, :] > standard_parameters[calculation_param]['extreme_upper_limit']))[0]
            calculation_arr[calculation_param_ii, out_of_bounds] = np.NaN

        #get indices for backscatter and scattering coefficient parameters
        backscat_param_index = [calculation_param_ii for calculation_param_ii, calculation_param in enumerate(calculation_params) if '_BACKSCAT_' in calculation_param][0]
        scat_param_index = [calculation_param_ii for calculation_param_ii, calculation_param in enumerate(calculation_params) if '_SCAT_' in calculation_param][0]

        #calculate new backscatter fraction
        total = calculation_arr[backscat_param_index,:] + calculation_arr[scat_param_index,:]
        bf = calculation_arr[backscat_param_index,:] / total

        return bf

    @staticmethod
    def calculate_asy(calculation_params, calculation_arr):
        """ Calculate asymmetry factor at a given wavelength

            :param calculation_params: GHOST names of the input components 
            :type calculation_params: list
            :param calculation_arr: input component values of shape (N components, time)
            :type calculation_arr: numpy.ndarray
            :return: new component array
            :rtype: numpy.ndarray
        """

        #filter out input data outside valid bounds
        for calculation_param_ii, calculation_param in enumerate(calculation_params):
            out_of_bounds = np.where(np.logical_or(calculation_arr[calculation_param_ii, :] < standard_parameters[calculation_param]['extreme_lower_limit'], calculation_arr[calculation_param_ii, :] > standard_parameters[calculation_param]['extreme_upper_limit']))[0]
            calculation_arr[calculation_param_ii, out_of_bounds] = np.NaN

        #calculate asymmetry factor 
        asy = -7.14*calculation_arr[0,:]**3 + 7.46*calculation_arr[0,:]**2 - 3.96*calculation_arr[0,:] + 0.9893

        return asy

    @staticmethod
    def calculate_ssa(calculation_params, calculation_arr):
        """ Calculate single scattering albedo (SSA) at a given wavelength

            :param calculation_params: GHOST names of the input components 
            :type calculation_params: list
            :param calculation_arr: input component values of shape (N components, time)
            :type calculation_arr: numpy.ndarray
            :return: new component array
            :rtype: numpy.ndarray
        """

        #filter out input data outside valid bounds
        for calculation_param_ii, calculation_param in enumerate(calculation_params):
            out_of_bounds = np.where(np.logical_or(calculation_arr[calculation_param_ii, :] < standard_parameters[calculation_param]['extreme_lower_limit'], calculation_arr[calculation_param_ii, :] > standard_parameters[calculation_param]['extreme_upper_limit']))[0]
            calculation_arr[calculation_param_ii, out_of_bounds] = np.NaN

        #get indices for absorption and extinction parameters
        abs_param_index = [calculation_param_ii for calculation_param_ii, calculation_param in enumerate(calculation_params) if 'ABSAOD_' in calculation_param][0]
        ext_param_index = [calculation_param_ii for calculation_param_ii, calculation_param in enumerate(calculation_params) if 'EXTAOD_' in calculation_param][0]

        #calculate new SSA component
        new_ssa = (calculation_arr[ext_param_index,:] - calculation_arr[abs_param_index,:]) / calculation_arr[ext_param_index,:]

        return new_ssa